#include "triangle.h"

void Triangle::Rasterize(FrameBuffer *fb, PPC *ppc, FrameBuffer *texture, 
  FrameBuffer *shadowMap, PPC *shadowMapPPC) {

  AABB bb;
  bb.AddPoint(verts[0]);
  bb.AddPoint(verts[1]);
  bb.AddPoint(verts[2]);

  if (!fb->ClipBB(bb))
    return;

  int lefti = (int) (bb[0][0]+0.5f);
  int righti = (int) (bb[1][0]-0.5f);
  int topi = (int) (bb[0][1]+0.5f);
  int bottomi = (int) (bb[1][1]-0.5f);
  M3x3 eeqs;
  eeqs[0] = SetEdgeEquation(0);
  eeqs[1] = SetEdgeEquation(1);
  eeqs[2] = SetEdgeEquation(2);

  M3x3 ssim;
  SetSSIM(ssim);
  M3x3 colorm;
  colorm[0] = colors[0];
  colorm[1] = colors[1];
  colorm[2] = colors[2];
  M3x3 colorABCs = (ssim*colorm).Transpose();
  V3 zABC = ssim*V3(verts[0][2], verts[1][2], verts[2][2]);
  V3 sABC = ssim*V3(tcs[0*2+0], tcs[1*2+0], tcs[2*2+0]);
  V3 tABC = ssim*V3(tcs[0*2+1], tcs[1*2+1], tcs[2*2+1]);

  for (int v = topi; v <= bottomi; v++) {
    for (int u = lefti; u <= righti; u++) {
      V3 pixv(.5f+(float)u, .5f+(float)v, 1.0f);
      V3 sdd = eeqs*pixv;
      if (sdd[0] < 0.0f || sdd[1] < 0.0f || 
        sdd[2] < 0.0f)
        continue;
//      unsigned int colui = color.GetColor();
      unsigned int colui = (colorABCs*pixv).GetColor();
      if (texture) {
        float currs = sABC*pixv;
        float currt = tABC*pixv;
        colui = texture->GetST(currs, currt);
      }
      float currz = zABC*pixv;
      if (!fb->IsCloser(currz, u, v))
        continue;
      fb->SetZB(currz, u, v);
      // shadow mapping
      if (shadowMap) {
        // unproject current pixel sample to 3D point
        V3 P = ppc->Unproject(u, v, currz);
        // project 3D point to shadow map PPC
        V3 projP;
        if (shadowMapPPC->Project(P, projP)) {
          // test with shadow map zbuffer
          if (!shadowMap->IsVisible(projP[2], (int) projP[0], (int) projP[1])) {
            colui = 0xFF000000;
          }
        }
      }
      fb->Set(u, v, colui);
    }
  }

}

V3 Triangle::SetEdgeEquation(int i) {

  V3 ret;
  V3 P0 = verts[i];
  V3 P1 = verts[(i+1)%3];
  ret[0] = P1[1]-P0[1];
  ret[1] = -P1[0]+P0[0];
  ret[2] = -P0[0]*ret[0]-P0[1]*ret[1];

  V3 P2 = verts[(i+2)%3];
  P2[2] = 1.0f;
  if (P2*ret < 0.0f)
    ret = ret * -1.0f;

  return ret;

}

void Triangle::DrawWireFrame(FrameBuffer *fb) {

  for (int i = 0; i < 3; i++) {
    int _i = (i+1)%3;
    fb->DrawSegment(verts[i], verts[_i], colors[i], colors[_i]);
  }

}

void Triangle::SetSSIM(M3x3 &ssim) {

  ssim[0] = V3(verts[0][0], verts[0][1], 1.0f);
  ssim[1] = V3(verts[1][0], verts[1][1], 1.0f);
  ssim[2] = V3(verts[2][0], verts[2][1], 1.0f);
  ssim = ssim.Invert();

}


void Triangle::RasterizePTM(FrameBuffer *fb, PPC *ppc, FrameBuffer *reffb, PPC *refppc) {


  // do not render if back facing
  V3 vs[3];
  for (int i = 0; i < 3; i++) {
    vs[i] = V3(verts[i][0], verts[i][1], 0.0f);
  }
  V3 bfn = (vs[1]-vs[0])^(vs[2]-vs[0]);
  if (bfn[2] < 0.0f)
    return;

  // clip projected triangle against image frame using axis aligned bounding box (2D)
  AABB bb;
  bb.AddPoint(verts[0]);
  bb.AddPoint(verts[1]);
  bb.AddPoint(verts[2]);
  if (!fb->ClipBB(bb))
    return;

  // retrieve clipped bounding box
  int lefti = (int) (bb[0][0]+0.5f);
  int righti = (int) (bb[1][0]-0.5f);
  int topi = (int) (bb[0][1]+0.5f);
  int bottomi = (int) (bb[1][1]-0.5f);

  // set up edge equations for rasterization
  M3x3 eeqs;
  eeqs[0] = SetEdgeEquation(0);
  eeqs[1] = SetEdgeEquation(1);
  eeqs[2] = SetEdgeEquation(2);

  // set up screen space interpolation matrix
  M3x3 ssim;
  SetSSIM(ssim);
  M3x3 colorm;
  colorm[0] = colors[0];
  colorm[1] = colors[1];
  colorm[2] = colors[2];
  M3x3 colorABCs = (ssim*colorm).Transpose();
  V3 zABC = ssim*V3(verts[0][2], verts[1][2], verts[2][2]);

  // for loops over bounding box
  for (int v = topi; v <= bottomi; v++) {
    for (int u = lefti; u <= righti; u++) {
      V3 pixv(.5f+(float)u, .5f+(float)v, 1.0f);
      V3 sdd = eeqs*pixv;
      if (sdd[0] < 0.0f || sdd[1] < 0.0f || 
        sdd[2] < 0.0f)
        continue; // pixel outside of triangle
      unsigned int colui = (colorABCs*pixv).GetColor();
      float currz = zABC*pixv;
      if (!fb->IsCloser(currz, u, v))
        continue; // sample farther than previously seen sample along this ray (i.e. this pixel location)

      // found sample inside triangle and visible
      // 1. unproject sample (i.e. find 3-D point corresponding to sample)
      V3 P = ppc->Unproject(u, v, currz);
      V3 projP;
      if (refppc->Project(P, projP)) {
        float refz = projP[2];
        int refu = (int) projP[0];
        int refv = (int) projP[1];
        if (!(refu < 0 || refu > reffb->w-1 || refv < 0 || refv > reffb->h-1)) {
          if (refz >= reffb->zb[(reffb->h-1-refv)*reffb->w+refu] - 1.0f)
            reffb->Get(projP[0], projP[1], colui);
        }
      }
      fb->SetZB(currz, u, v);
      fb->Set(u, v, colui);
    }
  }

}

